//    GeoExp is a 3D Studio Max plugin that export GEO files.
//    Copyright (C) 1999  Anders Reggestad (andersr@pvv.ntnu.no)
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
#include "geoout.h"

// Dump some global animation information.
void GeoOut::ExportGlobalInfo()
{
	fprintf(pStream,"GEO(Version=1)\n");
}

void GeoOut::ExportGeomObject(INode* node, int indentLevel)
{
	int	i,j;

	ObjectState os = node->EvalWorldState(GetStaticFrame());
	if (!os.obj || os.obj->SuperClassID()!=GEOMOBJECT_CLASS_ID) {
		return;
	}
	
	// Targets are actually geomobjects, but we will export them
	// from the camera and light objects, so we skip them here.
	if (os.obj->ClassID() == Class_ID(TARGET_CLASS_ID, 0))
		return;
	
	BOOL needDel = FALSE;
	TriObject* tri = GetTriObjectFromNode(node, GetStaticFrame(), needDel);
	if (!tri) {
		return;
	}
	
	Mesh* mesh = &tri->mesh;
	Matrix3 tm = node->GetObjTMAfterWSM(GetStaticFrame());
		
	fprintf(pStream,"Geometry(Name=\"%s\"){\n", FixupName(node->GetName()));

	///////////////////////////////////////////////////////////////
	// Vertex list

	fprintf(pStream,"  VertexList   = [ ");
	for (i=0; i<mesh->getNumVerts(); i++) {
		Point3 v = (tm * mesh->verts[i]);
		if (i != 0) fprintf(pStream,",\n                   ");
		fprintf(pStream, "( %s, %s, %s )",
			 Format(v.x), Format(v.y), Format(v.z));
	}
	fprintf(pStream," ];\n");

	///////////////////////////////////////////////////////////////
	// Normal list

	Face *face;;	
	Point3 *verts;
	Point3 v0, v1, v2;
	Tab<Point3> fnorms;
	Tab<VNormal> vnorms;

	// Compute face and vertex surface normals for the mesh
	face = mesh->faces;	
	verts = mesh->verts;
	fnorms.SetCount(mesh->getNumFaces());
	vnorms.SetCount(mesh->getNumVerts());

	for (i = 0; i < mesh->getNumVerts(); i++){
		vnorms[i] = VNormal();
	}
	for (i = 0; i < mesh->getNumFaces(); i++, face++) {
		v0 = verts[face->v[0]];
		v1 = verts[face->v[1]];
		v2 = verts[face->v[2]];
		fnorms[i] = (v1-v0)^(v2-v1);
		for (j = 0; j < 3; j++){
			vnorms[face->v[j]].AddNormal(fnorms[i],face->smGroup);
		}
		fnorms[i] = Normalize(fnorms[i]);
	}
	for (i = 0; i < mesh->getNumVerts(); i++){
		vnorms[i].Normalize();
	}


	if (vnorms.Count() > 0){

		Tab<Point3> norms;
		int numNorms = 0;

		// Count maximum number of normals
		for (i = 0; i < vnorms.Count(); i++) {
			VNormal * vnorm = vnorms.Addr(i);
			while (vnorm != NULL){
				numNorms++;
				vnorm = vnorm->next;
			}
		}
		norms.SetCount(numNorms);
		numNorms = 0;

		fprintf(pStream,"  NormalList   = [ ");
		for (i = 0; i < vnorms.Count(); i++) {
			VNormal * vnorm = vnorms.Addr(i);
			while (vnorm != NULL){
				int found = FALSE;
				j = 0;
				while (!found && (j < numNorms)){
					if (vnorm->norm == norms[j]){
						vnorm->index = j;
						found = TRUE;
						break;
					}
					j++;
				}
				if (!found){
					vnorm->index = numNorms++;
					norms[vnorm->index] = vnorm->norm;
					if (i != 0) fprintf(pStream,",\n                   ");
					fprintf(pStream,"( %s, %s, %s )",Format(vnorm->norm.x),
						Format(vnorm->norm.y), Format(vnorm->norm.z));
				}
				vnorm = vnorm->next;
			}
		}
		fprintf(pStream," ];\n");
	}

	///////////////////////////////////////////////////////////////
	// TexCoord list

	if (mesh->getNumTVerts() > 0){
		fprintf(pStream,"  TexCoordList = [ ");
		for (i=0; i<mesh->getNumTVerts(); i++) {
			Point3 uv = mesh->tVerts[i];
			if (i != 0) fprintf(pStream,",\n                   ");
			fprintf(pStream, "( %s, %s )",	Format(uv.x), Format(uv.y));
		}
		fprintf(pStream," ];\n");
	}
	

	///////////////////////////////////////////////////////////////
	// Polygon section

	Mtl* nodeMtl = node->GetMtl();
	if (nodeMtl && (nodeMtl->ClassID() == Class_ID(MULTI_CLASS_ID, 0)) ) {
		// Each polygon possibly have different materials 
		for (i=0; i<mesh->getNumFaces(); i++) {
			fprintf(pStream,"  Polygon {\n    Type = TRIANGLES;\n");

			Mtl* faceMtl = nodeMtl->GetSubMtl(mesh->faces[i].getMatID() % nodeMtl->NumSubMtls());
			if (faceMtl) {
				fprintf(pStream,"    Material = \"%s\";\n", faceMtl->GetName());
			}

			// Vertex
			fprintf(pStream,"    Vertex   = [ %d, %d, %d ];\n",
				mesh->faces[i].v[0], mesh->faces[i].v[1], mesh->faces[i].v[2]);

			// Normal
			VNormal *vn0,*vn1,*vn2;
			vn0 = vnorms.Addr(mesh->faces[i].v[0]);
			vn1 = vnorms.Addr(mesh->faces[i].v[1]);
			vn2 = vnorms.Addr(mesh->faces[i].v[2]);
			fprintf(pStream,"    Normal   = [ %d, %d, %d ];\n",vn0->GetIndex(mesh->faces[i].smGroup),
				vn1->GetIndex(mesh->faces[i].smGroup),vn2->GetIndex(mesh->faces[i].smGroup));

			// TexCoord
			if (mesh->getNumTVerts() > 0){
				fprintf(pStream,"    TexCoord = [ %d, %d, %d ];\n",
					mesh->tvFace[i].t[0], mesh->tvFace[i].t[1], mesh->tvFace[i].t[2]);
			}
			fprintf(pStream,"  }\n"); // Polygon
		}

	} else {
		// Every polygon either don't have material or have the same material
		fprintf(pStream,"  Polygon {\n    Type = TRIANGLES;\n");

		if (nodeMtl) fprintf(pStream,"    Material = \"%s\";\n", nodeMtl->GetName());

		// Vertex
		fprintf(pStream,"    Vertex   = [");
		for (i=0; i<mesh->getNumFaces(); i++) {
			if (i != 0) fprintf(pStream,",\n                ");
			fprintf(pStream," %d, %d, %d",mesh->faces[i].v[0],
				mesh->faces[i].v[1], mesh->faces[i].v[2]);
		}
		fprintf(pStream," ];\n");

		// Normal
		fprintf(pStream,"    Normal   = [");
		for (i=0; i<mesh->getNumFaces(); i++) {
			if (i != 0) fprintf(pStream,",\n                ");
			// Normal
			VNormal *vn0,*vn1,*vn2;
			vn0 = vnorms.Addr(mesh->faces[i].v[0]);
			vn1 = vnorms.Addr(mesh->faces[i].v[1]);
			vn2 = vnorms.Addr(mesh->faces[i].v[2]);
			fprintf(pStream," %d, %d, %d",vn0->GetIndex(mesh->faces[i].smGroup),
				vn1->GetIndex(mesh->faces[i].smGroup),vn2->GetIndex(mesh->faces[i].smGroup));
		}
		fprintf(pStream," ];\n");

		// TexCoord
		if (mesh->getNumTVerts() > 0){
			fprintf(pStream,"    TexCoord = [");
			for (i=0; i<mesh->getNumFaces(); i++) {
				if (i != 0) fprintf(pStream,",\n                ");
				fprintf(pStream," %d, %d, %d",mesh->tvFace[i].t[0],
					mesh->tvFace[i].t[1], mesh->tvFace[i].t[2]);
			}
			fprintf(pStream," ];\n"); // TexCoord
		}

		fprintf(pStream,"  }\n"); // Polygon

	}

	fprintf(pStream,"}\n");   // Geometry

	if (needDel) tri->DeleteMe();
}
